<?php

/* do NOT run this script through a web browser */
if (!isset($_SERVER['argv'][0]) || isset($_SERVER['REQUEST_METHOD'])  || isset($_SERVER['REMOTE_ADDR'])) {
  die("<br><strong>This script is only meant to run at the command line.</strong>");
}

$no_http_headers = true;

/* display no errors */
error_reporting(0);
//error_reporting(E_ALL);

require_once('teamspeak3_query_api.class.php');

// mail serverquery login cacti 4Sd7Np2h
// linux2           login cacti oYEI6tR8

/*
index: (no args)
  - get "serverlist" from server
  - get list of virtualserver_id
  - return the list

num_indexes: (no args)
  - do index
  - return count($idxs);

query: arg1=keyword
  print entries for keyword for all indexes by:
  - get cached "serverlist" from server
  - for all virtualserver_id from serverlist:
    - get cached "serverinfo" from server
    - print value for keyword

get: arg1=keyword, arg2=index
  return one entry for keyword with index (virtual server id) by:
  - get cached "serverinfo" from server with sid=index to $data
  - return the value for keyword $data[$keyword]

get cached "serverinfo" from server:
  - if last "serverinfo" is not valid any more:
    - send "serverinfo" to server with sid=$sid
    - for all keywords from serverinfo with numeric values:
      - write them to $_SESSION['teamspeak3'][$sid][$keyword]
  - return $_SESSION['teamspeak3'][$sid]

  cached is $_SESSION['teamspeak3'][$sid][$keyword] = $value
        and $_SESSION['teamspeak3_serverlist'][$sid][$keyword] = $value
*/

class ss_teamspeak3 extends teamspeak3_query_api {
  protected $debuglog;

  public function __construct($host, $port, $host_id, $user = '', $pass = '') {
    parent::__construct($host, $port, $user, $pass);

    ini_set('session.gc_probability', 0);
    session_save_path('/var/lib/php/session');
    session_id("cacti-ts3-{$host_id}-" . posix_getuid());
    session_start();
  }


  public function __destruct() {
    session_write_close();
    parent::__destruct();
  }


  public function debug_log_define($host_id) {
    $tmpdir = "/tmp/teamspeak-" . posix_getuid();
    is_dir($tmpdir) or mkdir($tmpdir, 0777, true);
    $this->debuglog = "{$tmpdir}/ts3-{$host_id}";
  }


  public function debug_log($msg) {
    if (!isset($this->debuglog))
      return;

    file_put_contents($this->debuglog, DateTime::createFromFormat('U.u', microtime(true))->format('Y-m-d H:i:s.u') . " {$msg}\n", FILE_APPEND);
  }


  public function debug_log_array($arr, $depth = 0, $prefix = '') {
    if (!isset($this->debuglog))
      return;

    if ($depth == 0) {
      $this->debug_log($prefix . ($prefix ? ' ' : '') . join(',', $arr));
    } else {
      foreach ($arr as $idx => $nested_arr) {
        $this->debug_log_array($nested_arr, $depth - 1, "{$prefix}[{$idx}]");
      }
    }
  }


  /* return timestamp of next poll according to $interval */
  public function next_poll($interval) {
    return floor(time() / $interval + 1) * $interval - 1;
  }


  /* combine connect and login to the server */
  public function enter_server() {

    if (isset($_SESSION['teamspeak3_blocked_until']) && $_SESSION['teamspeak3_blocked_until'] > time())
      throw new Exception('blocked');

    try {
      try {
        $this->is_connected() or $this->connect();
      }
      catch (Exception $e) {
        // retry connection once
        $this->is_connected() or $this->connect();
      }
      $this->is_logged_in() or strlen($this->user) and $this->login();
      unset($_SESSION['teamspeak3_blocked_until']);
    }
    catch (Exception $e) {
      // unable to connect or unable to login: don't access Teamspeak server
      // until next poll in 5 minutes (login failure) or 30 seconds (otherwise)
      // or exact seconds in case of a ban
      if (preg_match('/may retry in \d+ second/', $e->getMessage(), $matches)) {
        $_SESSION['teamspeak3_blocked_until'] = time() + $matches[1];
      } else {
        $_SESSION['teamspeak3_blocked_until'] = $this->next_poll($this->is_connected() ? 300 : 30);
      }
      throw new Exception("{$e->getMessage()} -> suppressing login until " . date('Y-m-d H:i:s', $_SESSION['teamspeak3_blocked_until']));
    }
  }


  /* perform serverlist and return list of virtual servers */
  public function get_serverlist() {

    // get data from server, if cached data is not valid
    if (!isset($_SESSION['teamspeak3_serverlist_valid']) || $_SESSION['teamspeak3_serverlist_valid'] <= time()) {

      $this->enter_server();

      // try to select the default server to get groups giving permission to call serverlist
      try {
        $this->select_virtualserver(array('port' => 9987));
      }
      catch (Exception $e) { } // ignore errors for this call

      $result = $this->serverlist();
      $data = array();
      array_map(function($e) use (&$data) { $data[$e['virtualserver_id']] = $e; }, $result);
      $_SESSION['teamspeak3_serverlist'] = $data;
      $_SESSION['teamspeak3_serverlist_valid'] = $this->next_poll(300);
    }

    $this->debug_log_array($_SESSION['teamspeak3_serverlist'], 1, 'serverlist');
    return $_SESSION['teamspeak3_serverlist'];
  }


  /* perform serverlist and return list of virtual server id's */
  public function get_serverlist_index() {
    return array_keys($this->get_serverlist());
  }


  /* perform serverinfo and return list of values for given server id */
  public function get_serverinfo($sid) {

    // get data from server, if cached data is not valid
    if (!isset($_SESSION['teamspeak3_valid_until'][$sid]) || $_SESSION['teamspeak3_valid_until'][$sid] <= time()) {

      $this->enter_server();
      $this->select_virtualserver(array('sid' => $sid));
      $result = $this->serverinfo();

      $result['virtualserver_voiceclientsonline'] = $result['virtualserver_clientsonline'] - $result['virtualserver_queryclientsonline'];

      // remove non-numeric values; they clutter the cache
      foreach ($result as $kw => &$value) {
        if (!is_numeric($value) && $kw != 'virtualserver_name')
          unset($result[$kw]);
      }

      $_SESSION['teamspeak3'][$sid] = $result;
      $_SESSION['teamspeak3_valid_until'][$sid] = $this->next_poll(300);
    }

    $this->debug_log_array($_SESSION['teamspeak3'][$sid], 0, "serverinfo[{$sid}]");
    return $_SESSION['teamspeak3'][$sid];
  }


  /* perform serverinfo and return requested value for given server id */
  public function get_serverinfo_value($sid, $kw) {

    // regular numeric output value
    $data = $this->get_serverinfo($sid);
    if (array_key_exists($kw, $data))
      return $data[$kw];

    // some additional values can be acquired with serverlist
    $data = $this->get_serverlist();
    return $data[$sid][$kw];
  }

}


/* ------------------------------------------------------------------------- */

if (!isset($called_by_script_server)) {
  include_once(dirname(__FILE__) . "/../include/global.php");
  array_shift($_SERVER['argv']);
  print call_user_func_array('ss_teamspeak3_info', $_SERVER['argv']);
}

function ss_teamspeak3_info($hostname, $host_id, $auth, $cmd, $arg1 = '', $arg2 = '') {

  if (!$cmd)
    return 'ERROR: usage: ss_teamspeak3_info(<hostname>, <host_id>, <auth>, <cmd>[, <kw>[, <idx>]])';

  // authentication info as username?password or port:username?password
  if (!preg_match('/^\s*(?:(?P<port>[\d]+):)?(?P<user>[^?]*)\?(?P<pass>\S*)\s*$/', $auth, $a))
    $a = array('port' => '', 'user' => '', 'pass' => '');

  $ts3 = new ss_teamspeak3($hostname, is_numeric($a['port']) ? $a['port'] : 10011, $host_id, $a['user'], $a['pass']);

  // uncomment if you want to see what data was retrieved
  // $ts3->debug_log_define($host_id);

  try {
    switch ($cmd) {
      case 'index':
        $idxs = $ts3->get_serverlist_index();
        foreach ($idxs as $sid) {
          print "{$sid}\n";
        }
        break;
      case 'num_indexes':
        $idxs = $ts3->get_serverlist_index();
        return count($idxs);
        break;
      case 'query':
        $kw = $arg1;
        $idxs = $ts3->get_serverlist_index();
        foreach ($idxs as $sid) {
          $value = $ts3->get_serverinfo_value($sid, $kw);
          print "{$sid}!{$value}\n";
        }
        break;
      case 'get':
        $kw = $arg1;
        $sid = $arg2;
        $value = $ts3->get_serverinfo_value($sid, $kw);
        return $value;
        break;
    }
  }

  catch (Exception $e) {
    if (isset($GLOBALS['called_by_script_server'])) {
      if ($e->getMessage() != 'blocked')
        cacti_log("ERROR: {$e->getMessage()}");
    } else {
      return "ERROR: {$e->getMessage()}\n";
    }
    return 'U';
  }
}
